const constant = {
    farmingMap: {
        farmland: ["grass_path", "grass", "dirt"],
        grass_path: ["dirt_with_roots", "mycelium", "podzol"]
    },
    event: {
        blocks: JSON.stringify(new Array)
    }
};
Object.defineProperty(constant, "NAMESPACE", {
    configurable: false,
    value: "VillagerMemory:",
    writable: false
});
Object.defineProperty(constant, "IDENTIFIER", {
    configurable: false,
    value: constant.NAMESPACE + "memory",
    writable: false
});
Object.defineProperty(constant, "TAG", {
    configurable: false,
    value: constant.NAMESPACE + "id",
    writable: false
});
Object.defineProperty(constant, "EMPTY", {
    configurable: false,
    value: new String,
    writable: false
});
Object.defineProperty(constant, "DISTANCE", {
    configurable: false,
    value: 2,
    writable: false
});

Object.defineProperty(constant, "getEntityWrapper", {
    configurable: false,
    writable: false,
    value: function (system) {
        const CommandContext = class {
            constructor(entity) {
                Object.defineProperty(this, "initializing", {
                    configurable: true,
                    value: true,
                    writable: false
                });
                Object.defineProperty(this, "closed", {
                    configurable: true,
                    value: false,
                    writable: false
                });
                Object.defineProperty(this, "tag", {
                    configurable: false,
                    value: constant.TAG + Math.floor(Math.random() * 0x10000000).toString(16),
                    writable: false
                });
                Object.defineProperty(this, "ref", {
                    configurable: false,
                    writable: false,
                    value: entity
                })
            }
            close() {
                if (this.initializing) {
                    return;
                }
                if (!this.closed) {
                    Object.defineProperty(this, "closed", {
                        configurable: false,
                        value: true,
                        writable: false
                    });
                    const tags = system.getComponent(this.ref.entity, "minecraft:tag");
                    const i = tags.data.indexOf(this.tag);
                    if (i == -1) {
                        return;
                    }
                    tags.data[i] = constant.EMPTY;
                    system.applyComponentChanges(this.ref.entity, tags)
                }
            }
            execute(cmd, callback, argCallback) {
                if (this.closed) {
                    return;
                }

                const command = `/execute @e[ c=1, tag=\"${this.tag}\" ] ~ ~ ~ ${cmd}`;

                if (this.initializing) {
                    const tags = system.getComponent(this.ref.entity, "minecraft:tag");
                    tags.data.push(this.tag);
                    system.applyComponentChanges(this.ref.entity, tags);
                    Object.defineProperty(this, "initializing", {
                        writable: false,
                        configurable: false,
                        value: false
                    })
                }

                if ((typeof callback) == "function") {
                    if ((typeof argCallback) == "function") {
                        const arg = argCallback();
                        system.executeCommand(command, (result) => {
                            callback(result, arg)
                        })
                    } else {
                        system.executeCommand(command, (result) => {
                            callback(result)
                        })
                    }
                } else {
                    const cmd = system.createEventData("minecraft:execute_command");
                    cmd.data.command = command;
                    system.broadcastEvent("minecraft:execute_command", cmd)
                }
            }
        };

        return class {
            constructor(entity) {
                Object.defineProperty(this, "entity", {
                    configurable: false,
                    writable: false,
                    value: entity
                })
            }
            isNotVillager() {
                return !(this.entity.__identifier__ == "minecraft:villager_v2" && system.isValidEntity(this.entity));
            }
            removeAllMarkTags() {
                const tags = system.getComponent(this.entity, "minecraft:tag");
                for (let i in tags.data) {
                    if (tags.data[i].startsWith(constant.NAMESPACE)) {
                        tags.data[i] = constant.EMPTY
                    }
                }
                system.applyComponentChanges(this.entity, tags)
            }
            createCommandContext() {
                return new CommandContext(this);
            }
            execute(cmd, callback) {
                const ctx = new CommandContext(this);
                ctx.execute(cmd, (result) => {
                    ctx.close();
                    callback(result);
                })
            }
        };
    }
});


const system = server.registerSystem(0, 0);


system.initialize = function () {
    this.registerComponent(constant.IDENTIFIER, constant.event);
    Object.defineProperty(this, "query", {
        value: this.registerQuery(),
        configurable: false,
        writable: false
    });

    Object.defineProperty(this, "Entity", {
        value: constant.getEntityWrapper(this),
        configurable: false,
        writable: false
    });

    this.listenForEvent("minecraft:entity_created", (event) => {
        const entity = new (this.Entity)(event.data.entity);

        if (entity.isNotVillager()) {
            return;
        }

        if (this.hasComponent(entity.entity, constant.IDENTIFIER)) {
            this.destroyComponent(entity.entity, constant.IDENTIFIER)
        }
        this.createComponent(entity.entity, constant.IDENTIFIER)

        entity.removeAllMarkTags()
    })
};

system.update = (function () {
    let tickTillUpdate = 0;

    return function () {
        if (tickTillUpdate == 0) {
            this.executeCommand("/gamerule randomtickspeed", (result) => {
                tickTillUpdate = Math.floor(20 / JSON.parse(result.data.details).randomtickspeed)
            });

            const entities = this.getEntitiesFromQuery(this.query);
            for (let entityFromAPI of entities) {
                const entity = new (this.Entity)(entityFromAPI);
                if (entity.isNotVillager() || !(this.hasComponent(entityFromAPI, constant.IDENTIFIER)) || Math.random() * 128 < 64) {
                    continue
                }

                const memory = this.getComponent(entityFromAPI, constant.IDENTIFIER);
                const blocks = new Set(JSON.parse(memory.data.blocks));

                const position = this.getComponent(entityFromAPI, "minecraft:position");
                for (let block of blocks) {
                    const [x, y, z] = block;
                    if ((Math.pow(x + 0.5 - position.data.x, 2) + Math.pow(y - position.data.y, 2) + Math.pow(z + 0.5 - position.data.z, 2)) < Math.pow(constant.DISTANCE * Math.SQRT2, 2)) {
                        entity.execute(`testforblock ${x} ${y} ${z} farmland`, (result) => {
                            if (result.data.statusCode != 0) {
                                const grass_path = () => {
                                    const ctx = entity.createCommandContext();
                                    let returning = constant.farmingMap.grass_path.length;
                                    let notFound = true;
                                    for (let targetBlock of constant.farmingMap.grass_path) {
                                        ctx.execute(`detect ${x} ${y} ${z} ${targetBlock} -1 setblock ${x} ${y} ${z} grass_path`, (result) => {
                                            returning--;
                                            if (result.data.statusCode == 0) {
                                                notFound = false;
                                                ctx.execute(`tp @s ~ ~ ~ facing ${x} ${y} ${z}`);
                                                ctx.execute("replaceitem entity @s slot.weapon.mainhand 0 iron_shovel")

                                            }
                                            if (returning == 0) {
                                                ctx.close();
                                                if (notFound) {
                                                    blocks.delete(block);
                                                    memory.data.blocks = JSON.stringify(Array.from(blocks));
                                                    this.applyComponentChanges(entityFromAPI, memory)
                                                }
                                            }
                                        })
                                    }
                                };

                                const ctx = entity.createCommandContext();
                                let returning = constant.farmingMap.farmland.length;
                                let notFound = true;
                                for (let targetBlock of constant.farmingMap.farmland) {
                                    ctx.execute(`detect ${x} ${y} ${z} ${targetBlock} -1 setblock ${x} ${y} ${z} farmland`, (result) => {
                                        returning--;
                                        if (result.data.statusCode == 0) {
                                            notFound = false;
                                            ctx.execute(`tp @s ~ ~ ~ facing ${x} ${y} ${z}`);
                                            ctx.execute("replaceitem entity @s slot.weapon.mainhand 0 iron_hoe")
                                        }
                                        if (returning == 0) {
                                            ctx.close();
                                            if (notFound) {
                                                grass_path()
                                            }
                                        }
                                    })
                                }
                            }
                        })
                    }
                }

                const ctx = entity.createCommandContext();
                let returning = 0;
                let added = false;
                for (let i = 0 - constant.DISTANCE; i != constant.DISTANCE + 1; i++) {
                    const x = ((i == 0) ? constant.EMPTY : i);
                    for (let j = 0 - constant.DISTANCE; j != constant.DISTANCE + 1; j++) {
                        returning++;
                        ctx.execute(`testforblock ~${x} ~-1 ~${(j == 0) ? constant.EMPTY : j} farmland`, (result, block) => {
                            returning--;

                            if (returning == 0) {
                                ctx.close()
                            }

                            if (result.data.statusCode == 0) {
                                block[0] += i;
                                block[1]--;
                                block[2] += j;
                                let pushing = true;
                                for (let pos of blocks) {
                                    if (pos[0] == block[0] && pos[1] == block[1] && pos[2] == block[2]) {
                                        pushing = false;
                                        break
                                    }
                                }
                                if (pushing) {
                                    added = true;
                                    blocks.add(block)
                                }
                            }
                            if (added && returning == 0) {
                                memory.data.blocks = JSON.stringify(Array.from(blocks));
                                this.applyComponentChanges(entityFromAPI, memory)
                            }
                        }, () => {
                            const arg = new Array(3);
                            const position = this.getComponent(entityFromAPI, "minecraft:position");
                            [arg[0], arg[1], arg[2]] = [Math.floor(position.data.x), Math.floor(position.data.y), Math.floor(position.data.z)];
                            return arg;
                        });
                    }
                }
            }
        } else {
            tickTillUpdate--
        }
    };
})();
